"""
Computes the "combined diff format" used by git.

This is one way of showing a 3+ way diff, such as when you have a tool that
needs to show multiple diffs to a common ancestor (like input, expected
output, and actual output); or a merge that combines multiple heads into one
common descendant.  These are not intended to be applied.

Instead of a single +/- character, there are n-1 of those characters at the
start of the line, depending on which of the outputs (or inputs, in the case of
a merge) that line is present in.

If you're ever in a situation where you're analyzing a diff of diffs, this
might be helpful in reducing confusion.

Documented at https://git-scm.com/docs/diff-format#_combined_diff_format

Example of two diffs vs a common ancestor:

    $ python -m moreorless.combined a b c
    --- a/file
    +++ b/file
    +++ b/file
    @@@ -1,2 +1,2 +1,2 @@@
      def main():
    --    pass
    +     print("b")
     +    print("hi")
"""

from difflib import SequenceMatcher
from typing import Dict, Iterable, List, Optional, Sequence, Union

from .combined_types import Hunk, Line, Snip


def combined_diff(
    from_files: List[str],
    to_files: List[str],
    basename: str = "file",
    from_filenames: Optional[List[str]] = None,
    to_filenames: Optional[List[str]] = None,
    context: int = 3,
) -> str:
    """
    Returns a combined diff representing a multiway change.
    """
    if not from_filenames:
        from_filenames = [f"a/{basename}"] * len(from_files)
    if not to_filenames:
        to_filenames = [f"b/{basename}"] * len(to_files)

    if len(from_files) != len(from_filenames):
        raise ValueError("Mismatched from_filenames length")
    if len(to_files) != len(to_filenames):
        raise ValueError("Mismatched to_filenames length")
    if len(from_files) != 1 and len(to_files) != 1:
        raise ValueError(
            f"one of from_files={len(from_files)} or to_files={len(to_files)} must be 1"
        )
    merge_mode = len(to_files) == 1

    file_header_symbols = ["-"] * len(from_files) + ["+"] * len(to_files)

    buf: List[str] = []
    for sym, fn in zip(file_header_symbols, from_filenames + to_filenames):
        buf.append(f"{sym}{sym}{sym} {fn}\n")

    for block in _group(
        _contributions(from_files, to_files, merge_mode=merge_mode), context=context
    ):
        if isinstance(block, Hunk):
            block.to_buf(buf, merge_mode)

    return "".join(buf)


def _contributions(
    from_files: Sequence[str],
    to_files: Sequence[str],
    merge_mode: bool,
) -> Sequence[Line]:
    """
    Calculate the contribution of various lines to the final result.

    Instead of storing the +/- symbols directly, we will store 1 in the
    column if the line exists in that file.

    This intermediate form is very simple to look for snip points (more than
    2*context runs of `(1, ...)`), and has the nice property that once
    you transpose to get columns, the `@@` lines for indices `i:j` start at
    `sum(col[:i])` and are `sum(col[i:j])` long.
    """
    if merge_mode:
        if len(to_files) != 1:
            raise ValueError(f"Can't merge_mode=True with to_files={len(to_files)}")
        common = to_files[0]
        rest = from_files
    else:
        if len(from_files) != 1:
            raise ValueError(f"Can't merge_mode=False with from_files={len(from_files)}")
        common = from_files[0]
        rest = to_files

    common_template = [0] * len(rest) + [1]

    common_lines = common.splitlines(True)
    common_contributions = [common_template.copy() for _ in common_lines]

    # dest_idx: {text: {idx: one_or_zero}}
    inserted_lines: List[Dict[str, Dict[int, int]]] = [
        {} for _ in range(len(common_lines) + 1)
    ]

    def helper(s: str, dest_idx: int) -> Dict[int, int]:
        if s not in inserted_lines[dest_idx]:
            inserted_lines[dest_idx][s] = {}
        return inserted_lines[dest_idx][s]

    for si, src in enumerate(rest):
        src_lines = src.splitlines(True)
        for tag, i1, i2, j1, j2 in SequenceMatcher(
            None, common_lines, src_lines
        ).get_opcodes():
            for i in range(i1, i2):
                if tag == "equal":
                    common_contributions[i][si] = 1

            for i in range(j1, j2):
                if tag == "insert":
                    helper(src_lines[i], i1)[si] = 1
                elif tag == "replace":
                    helper(src_lines[i], i1)[si] = 1

    result: List[Line] = []
    if merge_mode:
        for i in range(len(common_lines) + 1):
            for t, counts in inserted_lines[i].items():
                lst = [0] * len(rest)
                for a, b in counts.items():
                    lst[a] = b
                result.append(Line(tuple(lst) + (0,), t))

            if i < len(common_lines):
                result.append(Line(tuple(common_contributions[i]), common_lines[i]))
    else:
        for i in range(len(common_lines) + 1):
            if i < len(common_lines):
                result.append(Line(tuple(common_contributions[i]), common_lines[i]))

            for t, counts in inserted_lines[i].items():
                lst = [0] * len(rest)
                for a, b in counts.items():
                    lst[a] = b
                result.append(Line(tuple(lst) + (0,), t))

    return result


def _group(lines: Sequence[Line], context: int = 3) -> Iterable[Union[Hunk, Snip]]:

    if not lines:
        return

    no_change_count = [0] * len(lines)
    # Take as input the counts which represent symbols
    # + a
    #   b
    #   c
    #   d
    # + e
    for i, line in enumerate(lines):
        if line.is_context:
            no_change_count[i] = 1
    # Now we have a local which is 1 wherever it's common context
    # 0 + a
    # 1   b
    # 1   c
    # 1   d
    # 0 + e
    tmp_fwd = no_change_count.copy()
    tmp_rev = no_change_count.copy()

    for i in range(len(lines) - 1):
        if no_change_count[i] > 0 and no_change_count[i + 1] > 0:
            tmp_fwd[i + 1] = tmp_fwd[i] + 1

    for i in range(len(lines) - 1, 0, -1):
        if no_change_count[i] > 0 and no_change_count[i - 1] > 0:
            tmp_rev[i - 1] = tmp_rev[i] + 1

    result = [min(a, b) for a, b in zip(tmp_fwd, tmp_rev)]

    # Now we have the triangular count, and anywhere the count is >= context we can split
    # 0 + a
    # 1   b
    # 2   c
    # 1   d
    # 0 + e

    context_flag = 0
    buf = []
    omit_count = 0
    starts = [1] * len(line.contributions)

    for i, line in enumerate(lines):
        if result[i] <= context:
            if omit_count:
                yield Snip(omit_count)
            omit_count = 0
            context_flag = 0
            buf.append(line)
        elif context_flag:
            omit_count += 1
            starts = [x + y for x, y in zip(starts, line.contributions)]
        else:
            if buf and not all(x.is_context for x in buf):
                hunk = Hunk(buf, starts)
                yield hunk
                starts = [x + y for x, y in zip(starts, hunk.lengths)]
            buf = []
            context_flag = 1

    if buf and not all(x.is_context for x in buf):
        yield Hunk(buf, starts)


if __name__ == "__main__":  # pragma: no cover
    import sys
    from pathlib import Path

    lst = [Path(f).read_text() for f in sys.argv[1:]]
    print(combined_diff(lst[:1], lst[1:]), end="")



import sys
from pathlib import Path

import click

from . import unified_diff


def echo_color_unified_diff(astr: str, bstr: str, filename: str, n: int = 3) -> None:
    """
    Just like `moreorless.unified_diff` except using `click.secho`.
    """
    echo_color_precomputed_diff(unified_diff(astr, bstr, filename, n))


def echo_color_precomputed_diff(diff: str) -> None:
    """
    Like `echo_color_unified_diff`, but for precomputed diff results.
    """
    for line in diff.splitlines(True):
        # TODO benchmark and see if constructing the string up front is faster
        if line.startswith("---") or line.startswith("+++"):
            click.secho(line, bold=True, nl=False)
        elif line.startswith("@@"):
            click.secho(line, fg="cyan", nl=False)
        elif line.startswith("-"):
            click.secho(line, fg="red", nl=False)
        elif line.startswith("+"):
            click.secho(line, fg="green", nl=False)
        else:
            click.secho(line, nl=False)


def main(afile: str, bfile: str) -> None:  # pragma: no cover
    echo_color_unified_diff(Path(afile).read_text(), Path(bfile).read_text(), afile)


if __name__ == "__main__":  # pragma: no cover
    main(*sys.argv[1:])



import logging
import re
from dataclasses import dataclass, field
from typing import List, Optional, Sequence

__all__ = ["apply_single_file", "PatchException"]

LOG = logging.getLogger(__name__)


def apply_single_file(contents: str, patch: str, allow_offsets: bool = True) -> str:
    """
    Apply a clean patch, no fuzz, no rejects.
    """

    lines = contents.splitlines(True)
    hunks = _split_hunks(patch.splitlines(True)[2:])
    return "".join(_apply_hunks(lines, hunks, allow_offsets))


POSITION_LINE_RE = re.compile(r"@@ -(\d+)(?:,(\d+))? \+(\d+)(?:,(\d+))? @@[^\n]*\n?")


class PatchException(Exception):
    pass


class ContextException(PatchException):
    pass


def _parse_position_line(position_line: str) -> List[int]:
    """Given an `@@` line, return the four numbers within."""
    match = POSITION_LINE_RE.fullmatch(position_line)
    if not match:
        raise PatchException(f"Position line {position_line!r} failed to parse")
    return [
        int(match.group(1)),
        int(match.group(2) or "1"),
        int(match.group(3)),
        int(match.group(4) or "1"),
    ]


# TODO store the offsets too, to make filtering easier
@dataclass
class Hunk:
    position: Optional[List[int]] = None
    lines: List[str] = field(default_factory=list)


def _validate_hunk_counts(hunk: Hunk) -> None:
    pos = hunk.position
    assert pos is not None
    body = [line for line in hunk.lines[1:] if not line.startswith("\\")]
    old_count = sum(1 for line in body if line and line[0] in (" ", "-"))
    new_count = sum(1 for line in body if line and line[0] in (" ", "+"))
    if old_count != pos[1]:
        raise PatchException(
            f"Hunk header declares {pos[1]} old lines but body has {old_count}"
        )
    if new_count != pos[3]:
        raise PatchException(
            f"Hunk header declares {pos[3]} new lines but body has {new_count}"
        )


def _split_hunks(diff_lines: Sequence[str]) -> List[Hunk]:
    """
    Splits unified diff lines (after the file header) into hunks.
    """
    hunks: List[Hunk] = []
    hunk: Optional[Hunk] = None

    for line in diff_lines:
        if line.startswith("@@"):
            # Start a new hunk
            if hunk:
                _validate_hunk_counts(hunk)
                hunks.append(hunk)
            hunk = Hunk(_parse_position_line(line))
        # There should not be '---' or '+++' lines here, they are stripped off
        # in apply_single_file.
        if not hunk:
            raise PatchException(f"Lines without hunk header at {line!r}")
        hunk.lines.append(line)

    if hunk and hunk.lines:
        _validate_hunk_counts(hunk)
        hunks.append(hunk)

    return hunks


def _apply_hunks(lines: List[str], hunks: List[Hunk], allow_offsets: bool) -> List[str]:
    work = lines[:]
    file_offset = 0  # accumulation of delta
    prev_line = 0
    for hunk in hunks:
        assert hunk.position is not None
        pos = hunk.position[:]

        # If length is zero, this is a no-context deletion and per
        # https://www.artima.com/weblogs/viewpost.jsp?thread=164293 the numbers
        # are off by one from being actual line numbers. :/
        if pos[3] == 0:
            pos[2] += 1
        if pos[1] == 0:
            pos[0] += 1

        cur_line = pos[0] + file_offset - 1
        # Meld "No newline at end of file" up a line
        tmp = hunk.lines[:]
        for i in range(len(tmp) - 1, 0, -1):
            if tmp[i].startswith("\\ No newline"):
                del tmp[i]
                # strips newline (including dos newlines, although we don't
                # produce a those in moreorless.unified_diff)
                if tmp[i - 1].endswith("\r\n"):
                    tmp[i - 1] = tmp[i - 1][:-2]
                else:
                    tmp[i - 1] = tmp[i - 1][:-1]
        if allow_offsets:
            tmp2 = [t[1:] for t in tmp if t[0] in (" ", "-")]
            # TODO if hunks overlap, this checks against the already-modified
            # one for context, which seems wrong.  Unmodified file is something
            # like _context_match(lines, tmp2, ..., prev_line+file_offset)-file_offset

            # On a proper patch this always takes in cur_line and returns cur_line
            new_line = _context_match(work, tmp2, prev_line, len(work), cur_line)
            if new_line is None:
                raise PatchException(f"Failed to apply with offset at {cur_line}")
            if cur_line != new_line:
                LOG.info(f"Offset {new_line - cur_line}")
                cur_line = new_line

        for line in tmp[1:]:
            if line.startswith("-"):
                if line[1:] != work[cur_line]:
                    raise PatchException(f"DELETE fail at {cur_line}")
                del work[cur_line]
            elif line.startswith("+"):
                work.insert(cur_line, line[1:])
                cur_line += 1
            elif line.startswith(" "):
                if line[1:] != work[cur_line]:
                    raise PatchException(f"EQUAL fail at {cur_line}")
                cur_line += 1
            elif line.startswith("?"):  # pragma: no cover
                pass  # human readable line
            else:
                raise PatchException(f"Unknown line {line!r} at {cur_line}")
        file_offset += pos[3] - pos[1]
        prev_line = cur_line

    return work


def _context_match(
    file_lines: List[str],
    context_lines: List[str],
    range_start: int,
    range_end: int,
    start: int,
) -> Optional[int]:
    """
    Finds an offset within file_lines to match context.

    Returns i such that:
    * file_lines[i:i+len] == context_lines
    * i >= range_start
    * i <= range_end - len
    * minimizes abs(i-start)
    * minimizes i if there's a tie on abs
    """
    cl = len(context_lines)
    if not range_start >= 0:
        raise ContextException("context error 1: negative range_start")
    if not range_end >= range_start:
        raise ContextException("context error 2: flipped range")
    if not range_end <= len(file_lines):
        raise ContextException("context error 3: past end")
    if not start >= range_start:
        raise ContextException("context error 4: start before range_start")
    if not start <= range_end - cl:
        raise ContextException("context error 5: start past range_end")

    for di in range(0, max(start - range_start + 1, range_end - start - cl + 1)):
        t1 = start - di
        t2 = start + di
        if t1 >= range_start:
            if all(context_lines[j] == file_lines[t1 + j] for j in range(cl)):
                return t1
        if t2 + cl <= range_end:
            if all(context_lines[j] == file_lines[t2 + j] for j in range(cl)):
                return t2
    return None



from dataclasses import dataclass
from typing import Iterable, List, Sequence


@dataclass
class Line:
    # By convention, the `common` entry comes last (and typically isn't output
    # explicitly)
    contributions: Sequence[int]
    line: str

    @property
    def is_context(self) -> bool:
        return sum(self.contributions) == len(self.contributions)

    # @property
    # def is_common_only(self) -> bool:
    #     return self.contributions[-1] and sum(self.contributions) == 1

    def to_buf(self, buf: List[str], is_merge: bool) -> None:
        merge_mult = 1 if is_merge else -1

        if self.is_context:
            buf.append(" " * (len(self.contributions) - 1) + self.line)
        else:
            d = {0: " ", 1: "+", -1: "-"}
            for c in self.contributions[:-1]:
                buf.append(d[(self.contributions[-1] - c) * merge_mult])
            buf.append(self.line)
        if not self.line.endswith("\n"):
            buf.append("\n")
            buf.append("\\ No newline at end of file\n")

    def to_str(self, is_merge: bool) -> str:
        buf: List[str] = []
        self.to_buf(buf, is_merge)
        return "".join(buf)


@dataclass
class Hunk:
    lines: List[Line]
    starts: Sequence[int]
    lengths: Sequence[int] = ()

    def __post_init__(self) -> None:
        self.lengths = tuple([sum(col) for col in self._column_counts])

    @property
    def is_entirely_context(self) -> bool:
        return all(el.is_context for el in self.lines)

    @property
    def _column_counts(self) -> Iterable[Iterable[int]]:
        return zip(*[line.contributions for line in self.lines])

    def header(self, merge_mode: bool) -> str:
        buf: List[str] = []
        buf.append("@" * len(self.starts) + " ")
        if merge_mode:
            prefixes = ["-"] * (len(self.starts) - 1) + ["+"]
        else:
            prefixes = ["-"] + ["+"] * (len(self.starts) - 1)

        positions = list(zip(self.starts, self.lengths))
        if not merge_mode:
            positions.insert(0, positions.pop(-1))

        for p, x in zip(prefixes, positions):
            buf.append(f"{p}{x[0]},{x[1]} ")

        buf.append("@" * len(self.starts))
        buf.append("\n")
        return "".join(buf)

    def to_buf(self, buf: List[str], merge_mode: bool) -> None:
        if self.is_entirely_context:
            return

        buf.append(self.header(merge_mode))
        # print("S", self.starts)
        # print("L", self.lengths)
        for line in self.lines:
            # print("LINE", line.contributions)
            line.to_buf(buf, merge_mode)

    def to_str(self, is_merge: bool) -> str:
        buf: List[str] = []
        self.to_buf(buf, is_merge)
        return "".join(buf)


@dataclass
class Snip:
    count: int
